@@ -13,6 +13,7 @@
 #include "lib/ip.h"
 #include "lib/hash.h"
 #include "lib/resource.h"
+#include "lib/obstacle.h"
 #include "lib/timer.h"
 #include "lib/tlists.h"
 
@@ -28,25 +29,35 @@ struct config {
   list symbols;				/* Configured symbols in config order */
   TLIST_STRUCT_DEF(cli_config, struct cli_config) cli; /* Configured CLI sockets */
 
-  int mrtdump_file;			/* Configured MRTDump file (sysdep, fd in unix) */
+  struct rfile *mrtdump_file;		/* Configured MRTDump file */
   const char *syslog_name;		/* Name used for syslog (NULL -> no syslog) */
-  struct rtable_config *def_tables[NET_MAX]; /* Default routing tables for each network */
+  struct symbol *def_tables[NET_MAX];	/* Default routing tables for each network */
   struct iface_patt *router_id_from;	/* Configured list of router ID iface patterns */
 
   u32 router_id;			/* Our Router ID */
-  const char *hostname;			/* Hostname */
   u32 proto_default_debug;		/* Default protocol debug mask */
   u32 proto_default_mrtdump;		/* Default protocol mrtdump mask */
   u32 channel_default_debug;		/* Default channel debug mask */
   u32 table_default_debug;		/* Default table debug mask */
+  u32 show_route_debug;			/* Exports to CLI debug mask */
+  u16 filter_vstk, filter_estk;		/* Filter stack depth */
   struct timeformat tf_route;		/* Time format for 'show route' */
   struct timeformat tf_proto;		/* Time format for 'show protocol' */
   struct timeformat tf_log;		/* Time format for the logfile */
   struct timeformat tf_base;		/* Time format for other purposes */
   u32 gr_wait;				/* Graceful restart wait timeout (sec) */
+  const char *hostname;			/* Hostname */
 
   int cli_debug;			/* Tracing of CLI connections and commands */
-  int latency_debug;			/* I/O loop tracks duration of each event */
+  enum latency_debug_flags {
+    DL_PING = 1,
+    DL_WAKEUP = 2,
+    DL_SCHEDULING = 4,
+    DL_ALLOCATOR = 8,
+    DL_SOCKETS = 0x10,
+    DL_EVENTS = 0x20,
+    DL_TIMERS = 0x40,
+  } latency_debug;			/* I/O loops log information about task scheduling */
   u32 latency_limit;			/* Events with longer duration are logged (us) */
   u32 watchdog_warning;			/* I/O loop watchdog limit for warning (us) */
   u32 watchdog_timeout;			/* Watchdog timeout (in seconds, 0 = disabled) */
@@ -56,25 +67,58 @@ struct config {
   char *err_file_name;			/* File name containing error */
   char *file_name;			/* Name of main configuration file */
   int file_fd;				/* File descriptor of main configuration file */
+  int thread_count;			/* How many worker threads to prefork */
 
   struct sym_scope *root_scope;		/* Scope for root symbols */
   struct sym_scope *current_scope;	/* Current scope where we are actually in while parsing */
-  int obstacle_count;			/* Number of items blocking freeing of this config */
+  int allow_attributes;			/* Allow attributes in the current state of configuration parsing */
+  struct obstacle_target obstacles;	/* May be externally blocked */
+  struct callback obstacles_cleared;	/* Called when ready to delete */
   int shutdown;				/* This is a pseudo-config for daemon shutdown */
   int gr_down;				/* This is a pseudo-config for graceful restart */
-  btime load_time;			/* When we've got this configuration */
 };
 
+struct global_runtime {
+  struct timeformat tf_route;		/* Time format for 'show route' */
+  struct timeformat tf_proto;		/* Time format for 'show protocol' */
+  struct timeformat tf_log;		/* Time format for the logfile */
+  struct timeformat tf_base;		/* Time format for other purposes */
+
+  u32 gr_wait;				/* Graceful restart wait timeout (sec) */
+
+  u32 router_id;			/* Our Router ID */
+  const char *hostname;			/* Hostname */
+
+  btime load_time;			/* When we reconfigured last time */
+  int cli_debug;			/* Tracing of CLI connections and commands */
+  enum latency_debug_flags latency_debug;
+  u32 latency_limit;			/* Events with longer duration are logged (us) */
+  u32 watchdog_warning;			/* I/O loop watchdog limit for warning (us) */
+  u32 watchdog_timeout;			/* Watchdog timeout (in seconds, 0 = disabled) */
+};
+
+extern struct global_runtime * _Atomic global_runtime;
+
 /* Please don't use these variables in protocols. Use proto_config->global instead. */
-extern struct config *config;		/* Currently active configuration */
-extern struct config *new_config;	/* Configuration being parsed */
+typedef OBSREF(struct config) config_ref;
+
+/* Self-clearing local reference */
+static inline void config_ref_local_cleanup(config_ref *r)
+{ OBSREF_CLEAR(*r); }
+#define CONFIG_REF_LOCAL_EMPTY(_ref)				\
+  CLEANUP(config_ref_local_cleanup) config_ref _ref = {};
+#define CONFIG_REF_LOCAL(_ref, _val)				\
+  CONFIG_REF_LOCAL_EMPTY(_ref); OBSREF_SET(_ref, _val);
+
+extern config_ref config;		/* Currently active configuration */
+extern _Thread_local struct config *new_config;	/* Configuration being parsed */
 
 struct config *config_alloc(const char *name);
 int config_parse(struct config *);
-int cli_parse(struct config *);
+int cli_parse(struct config *, struct config *);
 void config_free(struct config *);
 void config_free_old(void);
-int config_commit(struct config *, int type, uint timeout);
+int config_commit(config_ref *, int type, uint timeout);
 int config_confirm(void);
 int config_undo(void);
 int config_status(void);
@@ -82,8 +126,6 @@ btime config_timer_status(void);
 void config_init(void);
 void cf_error(const char *msg, ...) NORET;
 #define cf_warn(msg, args...)  log(L_WARN "%s:%d:%d: " msg, ifs->file_name, ifs->lino, ifs->chno - ifs->toklen + 1, ##args)
-void config_add_obstacle(struct config *);
-void config_del_obstacle(struct config *);
 void order_shutdown(int gr);
 
 #define RECONFIG_NONE	0
@@ -131,7 +173,7 @@ struct symbol {
     const struct f_line *function;	/* For SYM_FUNCTION */
     const struct filter *filter;	/* For SYM_FILTER */
     struct rtable_config *table;	/* For SYM_TABLE */
-    struct f_dynamic_attr *attribute;	/* For SYM_ATTRIBUTE */
+    struct ea_class *attribute;		/* For SYM_ATTRIBUTE */
     struct mpls_domain_config *mpls_domain;	/* For SYM_MPLS_DOMAIN */
     struct mpls_range_config *mpls_range;	/* For SYM_MPLS_RANGE */
     struct f_val *val;			/* For SYM_CONSTANT */
@@ -151,12 +193,15 @@ struct sym_scope {
 
   uint slots;				/* Variable slots */
   byte soft_scopes;			/* Number of soft scopes above */
-  byte active:1;			/* Currently entered */
   byte block:1;				/* No independent stack frame */
   byte readonly:1;			/* Do not add new symbols */
 };
 
-extern struct sym_scope *global_root_scope;
+void cf_enter_filters(void);
+void cf_exit_filters(void);
+int cf_maybe_enter_filters(void);
+int cf_maybe_exit_filters(void);
+
 extern pool *global_root_scope_pool;
 extern linpool *global_root_scope_linpool;
 
@@ -203,7 +248,7 @@ struct include_file_stack {
 extern struct include_file_stack *ifs;
 
 int cf_lex(void);
-void cf_lex_init(int is_cli, struct config *c);
+void cf_lex_init(struct config *cli_main, struct config *c);
 void cf_lex_unwind(void);
 
 struct symbol *cf_find_symbol_scope(const struct sym_scope *scope, const byte *c);
@@ -224,6 +269,7 @@ static inline int cf_symbol_is_local(struct config *conf, struct symbol *sym)
 
 /* internal */
 struct symbol *cf_new_symbol(struct sym_scope *scope, pool *p, struct linpool *lp, const byte *c);
+struct symbol *cf_root_symbol(const byte *, struct sym_scope *);
 
 /**
  * cf_define_symbol - define meaning of a symbol
@@ -246,9 +292,6 @@ struct symbol *cf_new_symbol(struct sym_scope *scope, pool *p, struct linpool *l
     sym_->var_ = def_; \
     sym_; })
 
-#define cf_create_symbol(conf_, name_, type_, var_, def_) \
-  cf_define_symbol(conf_, cf_get_symbol(conf_, name_), type_, var_, def_)
-
 void cf_push_scope(struct config *, struct symbol *);
 void cf_pop_scope(struct config *);
 void cf_push_soft_scope(struct config *);
@@ -270,7 +313,7 @@ int cf_parse(void);
 /* Sysdep hooks */
 
 void sysdep_preconfig(struct config *);
-int sysdep_commit(struct config *, struct config *);
+void sysdep_commit(struct config *, struct config *);
 void sysdep_shutdown_done(void);
 
 #endif
